#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, List, cast
from web3.constants import CHECKSUM_ADDRESSS_ZERO

from skale.skale_base import SkaleBase
from skale.utils.contract_info import ContractInfo
from skale.utils.contract_types import ContractTypes
from skale.utils.helper import get_contracts_info

if TYPE_CHECKING:
    from eth_typing import ChecksumAddress
    from skale.contracts.allocator.allocator import Allocator


logger = logging.getLogger(__name__)


def spawn_skale_allocator_lib(skale: SkaleAllocator) -> SkaleAllocator:
    return SkaleAllocator(skale._endpoint, skale.instance.address, skale.wallet)


class SkaleAllocator(SkaleBase):
    """Represents skale-allocator smart contracts"""

    @property
    def project_name(self) -> str:
        return 'skale-allocator'

    @property
    def allocator(self) -> Allocator:
        return cast('Allocator', super()._get_contract('allocator'))

    def contracts_info(self) -> List[ContractInfo[SkaleAllocator]]:
        import skale.contracts.allocator as contracts

        return [
            ContractInfo('escrow', 'Escrow', contracts.Escrow, ContractTypes.API, True),
            ContractInfo('allocator', 'Allocator', contracts.Allocator, ContractTypes.API, True),
        ]

    def get_contract_address(self, name: str) -> ChecksumAddress:
        if name == 'Escrow':
            return CHECKSUM_ADDRESSS_ZERO
        return super().get_contract_address(name)

    def set_contracts_info(self) -> None:
        self._SkaleBase__contracts_info = get_contracts_info(self.contracts_info())



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations
import logging
from typing import List, TYPE_CHECKING, cast

from skale.skale_base import SkaleBase
from skale.utils.contract_info import ContractInfo
from skale.utils.contract_types import ContractTypes
from skale.utils.helper import get_contracts_info

if TYPE_CHECKING:
    import skale.contracts.manager as contracts


logger = logging.getLogger(__name__)


class SkaleManager(SkaleBase):
    """Represents skale-manager smart contracts"""

    @property
    def project_name(self) -> str:
        return 'skale-manager'

    @staticmethod
    def contracts_info() -> List[ContractInfo[SkaleManager]]:
        import skale.contracts.manager as contracts

        return [
            ContractInfo(
                'contract_manager',
                'ContractManager',
                contracts.ContractManager,
                ContractTypes.API,
                False,
            ),
            ContractInfo('token', 'SkaleToken', contracts.Token, ContractTypes.API, False),
            ContractInfo('manager', 'SkaleManager', contracts.Manager, ContractTypes.API, True),
            ContractInfo(
                'constants_holder',
                'ConstantsHolder',
                contracts.ConstantsHolder,
                ContractTypes.INTERNAL,
                True,
            ),
            ContractInfo('nodes', 'Nodes', contracts.Nodes, ContractTypes.API, True),
            ContractInfo(
                'node_rotation', 'NodeRotation', contracts.NodeRotation, ContractTypes.API, True
            ),
            ContractInfo('schains', 'Schains', contracts.SChains, ContractTypes.API, True),
            ContractInfo(
                'schains_internal',
                'SchainsInternal',
                contracts.SChainsInternal,
                ContractTypes.API,
                True,
            ),
            ContractInfo('dkg', 'SkaleDKG', contracts.DKG, ContractTypes.API, True),
            ContractInfo(
                'key_storage', 'KeyStorage', contracts.KeyStorage, ContractTypes.API, True
            ),
            ContractInfo(
                'delegation_controller',
                'DelegationController',
                contracts.DelegationController,
                ContractTypes.API,
                False,
            ),
            ContractInfo(
                'delegation_period_manager',
                'DelegationPeriodManager',
                contracts.DelegationPeriodManager,
                ContractTypes.API,
                False,
            ),
            ContractInfo(
                'validator_service',
                'ValidatorService',
                contracts.ValidatorService,
                ContractTypes.API,
                False,
            ),
            ContractInfo(
                'token_state', 'TokenState', contracts.TokenState, ContractTypes.API, False
            ),
            ContractInfo(
                'distributor', 'Distributor', contracts.Distributor, ContractTypes.API, False
            ),
            ContractInfo(
                'slashing_table', 'SlashingTable', contracts.SlashingTable, ContractTypes.API, False
            ),
            ContractInfo('wallets', 'Wallets', contracts.Wallets, ContractTypes.API, True),
            ContractInfo('bounty_v2', 'BountyV2', contracts.BountyV2, ContractTypes.API, True),
            ContractInfo('punisher', 'Punisher', contracts.Punisher, ContractTypes.API, True),
            ContractInfo(
                'sync_manager', 'SyncManager', contracts.SyncManager, ContractTypes.API, False
            ),
            ContractInfo(
                'time_helpers_with_debug',
                'TimeHelpersWithDebug',
                contracts.TimeHelpersWithDebug,
                ContractTypes.API,
                False,
            ),
        ]

    @property
    def bounty_v2(self) -> contracts.BountyV2:
        return cast('contracts.BountyV2', self._get_contract('bounty_v2'))

    @property
    def constants_holder(self) -> contracts.ConstantsHolder:
        return cast('contracts.ConstantsHolder', self._get_contract('constants_holder'))

    @property
    def contract_manager(self) -> contracts.ContractManager:
        return cast('contracts.ContractManager', self._get_contract('contract_manager'))

    @property
    def delegation_controller(self) -> contracts.DelegationController:
        return cast('contracts.DelegationController', self._get_contract('delegation_controller'))

    @property
    def delegation_period_manager(self) -> contracts.DelegationPeriodManager:
        return cast(
            'contracts.DelegationPeriodManager', self._get_contract('delegation_period_manager')
        )

    @property
    def distributor(self) -> contracts.Distributor:
        return cast('contracts.Distributor', self._get_contract('distributor'))

    @property
    def dkg(self) -> contracts.DKG:
        return cast('contracts.DKG', self._get_contract('dkg'))

    @property
    def key_storage(self) -> contracts.KeyStorage:
        return cast('contracts.KeyStorage', self._get_contract('key_storage'))

    @property
    def manager(self) -> contracts.Manager:
        return cast('contracts.Manager', self._get_contract('manager'))

    @property
    def node_rotation(self) -> contracts.NodeRotation:
        return cast('contracts.NodeRotation', self._get_contract('node_rotation'))

    @property
    def nodes(self) -> contracts.Nodes:
        return cast('contracts.Nodes', self._get_contract('nodes'))

    @property
    def punisher(self) -> contracts.Punisher:
        return cast('contracts.Punisher', self._get_contract('punisher'))

    @property
    def schains(self) -> contracts.SChains:
        return cast('contracts.SChains', self._get_contract('schains'))

    @property
    def schains_internal(self) -> contracts.SChainsInternal:
        return cast('contracts.SChainsInternal', self._get_contract('schains_internal'))

    @property
    def slashing_table(self) -> contracts.SlashingTable:
        return cast('contracts.SlashingTable', self._get_contract('slashing_table'))

    @property
    def sync_manager(self) -> contracts.SyncManager:
        return cast('contracts.SyncManager', self._get_contract('sync_manager'))

    @property
    def time_helpers_with_debug(self) -> contracts.TimeHelpersWithDebug:
        return cast('contracts.TimeHelpersWithDebug', self._get_contract('time_helpers_with_debug'))

    @property
    def token(self) -> contracts.Token:
        return cast('contracts.Token', self._get_contract('token'))

    @property
    def token_state(self) -> contracts.TokenState:
        return cast('contracts.TokenState', self._get_contract('token_state'))

    @property
    def validator_service(self) -> contracts.ValidatorService:
        return cast('contracts.ValidatorService', self._get_contract('validator_service'))

    @property
    def wallets(self) -> contracts.Wallets:
        return cast('contracts.Wallets', self._get_contract('wallets'))

    def init_contract_manager(self) -> None:
        from skale.contracts.manager.contract_manager import ContractManager

        self.add_lib_contract('contract_manager', ContractManager, 'ContractManager')

    def set_contracts_info(self) -> None:
        self.init_contract_manager()
        self._SkaleBase__contracts_info = get_contracts_info(self.contracts_info())


def spawn_skale_manager_lib(skale: SkaleManager) -> SkaleManager:
    """Clone skale manager object with the same wallet"""
    return SkaleManager(skale._endpoint, skale.instance.address, skale.wallet)



from skale.contracts.base_contract import BaseContract, transaction_method
from skale.contracts.manager.bounty_v2 import BountyV2
from skale.contracts.manager.constants_holder import ConstantsHolder
from skale.contracts.manager.contract_manager import ContractManager
from skale.contracts.manager.delegation.delegation_controller import DelegationController
from skale.contracts.manager.delegation.delegation_period_manager import DelegationPeriodManager
from skale.contracts.manager.delegation.distributor import Distributor
from skale.contracts.manager.delegation.slashing_table import SlashingTable
from skale.contracts.manager.delegation.token_state import TokenState
from skale.contracts.manager.delegation.validator_service import ValidatorService
from skale.contracts.manager.dkg import DKG
from skale.contracts.manager.groups import Groups
from skale.contracts.manager.key_storage import KeyStorage
from skale.contracts.manager.manager import Manager
from skale.contracts.manager.node_rotation import NodeRotation
from skale.contracts.manager.nodes import Nodes
from skale.contracts.manager.punisher import Punisher
from skale.contracts.manager.schains import SChains
from skale.contracts.manager.schains_internal import SChainsInternal
from skale.contracts.manager.sync_manager import SyncManager
from skale.contracts.manager.test.time_helpers_with_debug import TimeHelpersWithDebug
from skale.contracts.manager.token import Token
from skale.contracts.manager.wallets import Wallets

__all__ = [
    'BountyV2',
    'ConstantsHolder',
    'ContractManager',
    'DelegationController',
    'DelegationPeriodManager',
    'Distributor',
    'DKG',
    'KeyStorage',
    'Manager',
    'NodeRotation',
    'Nodes',
    'Punisher',
    'SChains',
    'SChainsInternal',
    'SlashingTable',
    'SyncManager',
    'TimeHelpersWithDebug',
    'Token',
    'TokenState',
    'ValidatorService',
    'Wallets',
    'BaseContract',
    'transaction_method',
    'Groups',
]



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import abc
import logging
from typing import TYPE_CHECKING, Dict, Self, Type

from skale_contracts import skale_contracts

from skale.utils.exceptions import InvalidWalletError, EmptyWalletError
from skale.utils.web3_utils import default_gas_price, init_web3
from skale.wallets import BaseWallet

if TYPE_CHECKING:
    from eth_typing import ChecksumAddress
    from skale.contracts.base_contract import BaseContract
    from skale.utils.contract_info import ContractInfo


logger = logging.getLogger(__name__)


class EmptyPrivateKey(Exception):
    pass


class SkaleBase:
    __metaclass__ = abc.ABCMeta

    def __init__(
        self,
        endpoint: str,
        alias_or_address: str,
        wallet: BaseWallet | None = None,
        state_path: str | None = None,
        ts_diff: int | None = None,
        provider_timeout: int = 30,
    ):
        logger.info(
            'Initializing skale.py, endpoint: %s, wallet: %s', endpoint, type(wallet).__name__
        )
        self._endpoint = endpoint
        self.web3 = init_web3(
            endpoint, state_path=state_path, ts_diff=ts_diff, provider_timeout=provider_timeout
        )
        self.network = skale_contracts.get_network_by_provider(self.web3.provider)
        self.project = self.network.get_project(self.project_name)
        self.instance = self.project.get_instance(alias_or_address)
        self.__contracts: Dict[str, BaseContract[Self]] = {}
        self.__contracts_info: Dict[str, ContractInfo[Self]] = {}
        self.set_contracts_info()
        if wallet:
            self.wallet = wallet

    @property
    @abc.abstractmethod
    def project_name(self) -> str:
        """Name of smart contracts project"""

    @property
    def gas_price(self) -> int:
        return default_gas_price(self.web3)

    @property
    def wallet(self) -> BaseWallet:
        if not self._wallet:
            raise EmptyWalletError('No wallet provided')
        return self._wallet

    @wallet.setter
    def wallet(self, wallet: BaseWallet) -> None:
        if issubclass(type(wallet), BaseWallet):
            self._wallet = wallet
        else:
            raise InvalidWalletError(
                f'Wrong wallet class: {type(wallet).__name__}. \
                                       Must be one of the BaseWallet subclasses'
            )

    @abc.abstractmethod
    def set_contracts_info(self) -> None:
        pass

    def __init_contract_from_info(self, contract_info: ContractInfo[Self]) -> None:
        if contract_info.upgradeable:
            self.init_upgradeable_contract(contract_info)
        else:
            self.add_lib_contract(
                contract_info.name, contract_info.contract_class, contract_info.contract_name
            )

    def init_upgradeable_contract(self, contract_info: ContractInfo[Self]) -> None:
        address = self.get_contract_address(contract_info.contract_name)
        self.add_lib_contract(
            contract_info.name, contract_info.contract_class, contract_info.contract_name, address
        )

    def add_lib_contract(
        self,
        name: str,
        contract_class: Type[BaseContract[Self]],
        contract_name: str,
        contract_address: ChecksumAddress | None = None,
    ) -> None:
        address = contract_address or self.instance.get_contract_address(contract_name)
        logger.debug('Fetching abi for %s, address %s', name, address)
        contract_abi = self.instance.abi[contract_name]
        self.add_contract(name, contract_class(self, name, address, contract_abi))

    def add_contract(self, name: str, contract: BaseContract[Self]) -> None:
        self.__contracts[name] = contract

    def get_contract_address(self, name: str) -> ChecksumAddress:
        return self.web3.to_checksum_address(self.instance.get_contract_address(name))

    def _get_contract(self, name: str) -> BaseContract[Self]:
        if name not in self.__contracts:
            if not self.__contracts_info.get(name):
                logger.warning("%s method/contract wasn't found", name)
                raise ValueError(name, ' is an unknown contract')
            logger.debug("Contract %s wasn't inited, creating now", name)
            contract_info = self.__contracts_info[name]
            self.__init_contract_from_info(contract_info)
        return self.__get_contract_by_name(name)

    def __get_contract_by_name(self, name: str) -> BaseContract[Self]:
        return self.__contracts[name]

    def __getattr__(self, name: str) -> BaseContract[Self]:
        return self._get_contract(name)
